#ifndef WIOS_H_
#define WIOS_H_
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <pthread.h>
#include <sndfile.h>
#include <limits.h>
#include <sys/time.h>


/* -------------------------------------------------- 
 * 以下は基本的に編集する必要はありません 
 * デフォルトの挙動を変えたい場合のみ編集してください
 * -------------------------------------------------- */

#define WIOS_ALL_SEC     0
#define WIOS_MIN_SEC     1



/* --------------------------------------------------
 * デフォルト設定値
 * -------------------------------------------------- */

/* -------------------------------------------------- 
 * ここから編集可能領域 */

#define DEFAULT_DAFILE_NAME        "da.wav"
#define DEFAULT_ADFILE_NAME        "ad.wav"
#define TDBD_DEFAULT_OCHFILE_NAME  "da.och"
#define TDBD_DEFAULT_ICHFILE_NAME  "ad.ich"

#define ALSA_DEFAULT_DEVICE_NAME   "plughw:0,0"
#define TDBD_DEFAULT_DEVICE_NAME   "/dev/sinichusb0"
#define RASP_DEFAULT_DEVICE_NAME   "192.168.33.24"

#define DEFAULT_EXEC_SEC            WIOS_ALL_SEC

#define DEFAULT_CHANNELS            1
#define DEFAULT_RATE                16000
#define DEFAULT_FORMAT              WIOS_PCM_FORMAT_S16
#define DEFAULT_ENCODING						"16"

#define DEFAULT_WIOS_WARN_LEVEL     1
#define DEFAULT_WIOS_VERBOS_LEVEL   1

#define DEFAULT_PERIODS             2
#define DEFAULT_PERIODSIZE          1024

/* ここまで編集可能領域
 * -------------------------------------------------- */



/* -------------------------------------------------- 
 * 以下は編集する必要はありません
 * -------------------------------------------------- */

#define WIOS_WARN_LEVEL_MIN         0
#define WIOS_WARN_LEVEL_MAX         2
#define WIOS_VERBOS_LEVEL_MIN       0
#define WIOS_VERBOS_LEVEL_MAX       2
#define MAX_STRING_LENGTH           64

#define DUMMY_BYTE        100

#define WIOS_SUCCESS        0
#define WIOS_EOF          (-1)

#define WIOS_TRUE           1
#define WIOS_FALSE          0

#define WIOS_VERSION     "0.1.3"



typedef int32_t int24_t;

typedef enum {
	WIOS_PCM_FORMAT_U8,
	WIOS_PCM_FORMAT_S16,
	WIOS_PCM_FORMAT_S24,
	WIOS_PCM_FORMAT_S32,
	WIOS_PCM_FORMAT_UNKNOWN
} wios_pcm_format_t;

typedef struct { /* ファイルフォーマット情報 */
	int               channels;
	int               rate;
	wios_pcm_format_t format;
} wios_fmtinfo_t;

typedef struct { /* 再録時に必要な情報 */
	int  periods;
	int  periodsize;
	int  sec;
} wios_execinfo_t;

typedef struct { /* 波形データに関する情報 */
	int              byte_per_item;
	int              all_frames;
	int              buf_size; /* periodsize*channels [item] */
	int              loops, loops_org;
	wios_execinfo_t  execinfo;
	double           sec_per_frame;
} wios_signal_t;

typedef struct { /* libsndfile用ファイルハンドル */
	SNDFILE  * sndfile;
	SF_INFO  sfinfo;
} lsf_t;

typedef enum {
	wios_da,
	wios_ad
} wios_io_type;

typedef struct {
	void         * fp;                /* ファイルハンドル */
	void         * dp;                /* デバイスハンドル */
	const char   * file_name;         /* ファイル名 */
	const char   * dev_name;          /* デバイス名 */
	void         * priv;              /* デバイスに固有な構造体を指す */
	lsf_t          sf;                /* libsndfile用ファイルハンドル */
	wios_fmtinfo_t fmtinfo;           /* フォーマット情報 */ 
	wios_signal_t  signal;            /* 波形データに関する情報 */
	wios_io_type   io_type;           /* D/A or A/D データ */
	int            clips;             /* クリッピング回数 */
	int            max_item_val;      /* 処理したアイテムの数値の中の最大値 */
	int            min_item_val;      /* 処理したアイテムの数値の中の最小値	*/
	long long      success_byte_cnt;  /* 再生(録音)できたバイト数 */
	long long      missing_byte_cnt;  /* 再生(録音)できなかったバイト数 */
	int            gain;              /* ゲインの値 */
} wios_file;

typedef wios_file wios_file_t;

typedef enum {
	wios_alsa, wios_tdbd, wios_rasp, wios_unknown
} wios_device_t;

/* 同期モードで異なるデバイスを使うときだけwios_da_dev, wios_ad_devを使う */
/* それ以外の時はwios_dev以外は使えないように実装してある */
wios_device_t wios_dev;
wios_device_t wios_da_dev, wios_ad_dev;

typedef struct { /* ファイルフォーマットハンドラ */
	int    (*device_init)(wios_file_t * ft);
	wios_file_t * (*open_read)(
		const char             * path,
		const char             * dev,
		const wios_execinfo_t  * info);
	wios_file_t * (*open_write)(
		const char             * path,
		const char             * dev,
		const wios_execinfo_t  * info,
		const wios_fmtinfo_t   * fmt);
	void * (*playback)(void * arg);
	void * (*capture)(void * arg);
	void   (*finish)(wios_file_t * ft);
} wios_format_handler_t;

typedef struct { /* 再録に使うバッファ */
	int32_t  * fl;
	char     * dv;
} wios_buffer_t;



/* GLOBAL VARIABLE */
/* -------------------------------------------------- */

int wios_warn_level;   /* ワーニング表示レベル */
int wios_verbos_level; /* メッセージ表示レベル */

enum {
	wios_play, wios_rec, wios_sync, wios_help
} wios_mode;

static wios_pcm_format_t wios_format[] = {
	WIOS_PCM_FORMAT_U8,
	WIOS_PCM_FORMAT_S16,
	WIOS_PCM_FORMAT_S24,
	WIOS_PCM_FORMAT_S32,
	WIOS_PCM_FORMAT_UNKNOWN
};

static char supported_encoding[][3] = {
	"u8",
	"16",
	"24",
	"32",
	"uk"
};

/* 同期モードで異なるデバイスを使うときだけda_handler, ad_handlerを使う */
/* それ以外の時はhandler以外は使えないように実装してある */
wios_format_handler_t const * handler;
wios_format_handler_t const * da_handler;
wios_format_handler_t const * ad_handler;

int              sync_start_flag;  /* 同期スタートフラグ */
pthread_mutex_t  ready_mutex;      /* 再録準備用mutex */
pthread_cond_t   ready_cond;       /* 再録準備用条件変数 */
pthread_mutex_t  sync_mutex;       /* 同期時にスレッドを待機させるmutex */
pthread_cond_t   sync_cond;        /* 同期事にスレッドを待機させる条件変数 */
pthread_t        da_thread;        /* 再生用スレッドID */
pthread_t        ad_thread;        /* 録音用スレッドID */

int playback_end_flag;  /* 再生完了を録音スレッドに知らせる */
int loop_break_flag;    /* 再録ループ強制終了フラグ */

int print_result_flag;  /* プログラムの結果を表示するかどうか */
int print_status_flag;  /* ファイル情報を表示するかどうか */
int print_mesg_flag;    /* 再録のタイミングを表示するかどうか */
int print_time_flag;    /* 再録処理開始の時刻を表示するかどうか */

int use_ch_option_flag;   /* チャンネル数が指定されたかどうか */
int use_enc_option_flag;  /* エンコーディングが指定されたかどうか */
int use_da_dev_option_flag; /* DAデバイス名が指定されたかどうか */
int use_ad_dev_option_flag; /* ADデバイス名が指定されたかどうか */
int use_2dev_option_flag; /* 同期モードで再録に違うデバイスが指定されたか */

char * myname;       /* プログラム名 */
wios_file_t * ftda;  /* 再生ファイルハンドル */
wios_file_t * ftad;  /* 録音ファイルハンドル */
char * da_path;      /* 再生ファイルパス */
char * ad_path;      /* 録音ファイルパス */
char * da_dev;       /* 再生デバイス名 */
char * ad_dev;       /* 録音デバイス名 */
wios_execinfo_t  execinfo;    /* 再録時に必要な情報 */
wios_fmtinfo_t   ad_fmtinfo;  /* 録音ファイルフォーマット */



// TEST CODE START
#define p(x) printf("%s:%d\n", #x, x)
#define pp(x,y) printf("%s:%d, %s:%d\n", #x, x, #y, y);
// TEST CODE END



#ifdef min
#undef min
#endif
#define min(a, b) ((a) <= (b) ? (a) : (b))

#ifdef max
#undef max
#endif
#define max(a, b) ((a) >= (b) ? (a) : (b))

#define WIOS_INT_MIN(bits) (1<<((bits)-1))
#define WIOS_INT_MAX(bits) (((unsigned)-1)>>(33-(bits)))
#define WIOS_UINT_MAX(bits) (WIOS_INT_MIN(bits)|WIOS_INT_MAX(bits))

#define WIOS_CONVERT_32BIT_TO_UNSIGNED(bits,d) (uint##bits##_t)(WIOS_CONVERT_32BIT_TO_SIGNED(bits,d)^WIOS_INT_MIN(bits))
#define WIOS_CONVERT_32BIT_TO_SIGNED(bits,d) (int##bits##_t)((d)>>(32-(bits)))

#define WIOS_CONVERT_32_TO_U8BIT(d) WIOS_CONVERT_32BIT_TO_UNSIGNED(8,d)
#define WIOS_CONVERT_32_TO_16BIT(d) WIOS_CONVERT_32BIT_TO_SIGNED(16,d)
#define WIOS_CONVERT_32_TO_24BIT(d) WIOS_CONVERT_32BIT_TO_SIGNED(24,d)

#define WIOS_SIGNED_CLIP_COUNT(bits,d,clips) ((((d) >= (int##bits##_t)WIOS_INT_MAX(bits)) || ((d) <= (bits==24?-1:1)*(int##bits##_t)WIOS_INT_MIN(bits))) ? (clips)++, WIOS_TRUE : WIOS_FALSE)
#define WIOS_UNSIGNED_CLIP_COUNT(bits,d,clips) ((((d) >= (uint##bits##_t)WIOS_UINT_MAX(bits)) || ((d) <= (uint##bits##_t)0)) ? (clips)++, WIOS_TRUE : WIOS_FALSE)

#define WIOS_U8BIT_CLIP_COUNT(d,clips) WIOS_UNSIGNED_CLIP_COUNT(8,d,clips)
#define WIOS_16BIT_CLIP_COUNT(d,clips) WIOS_SIGNED_CLIP_COUNT(16,d,clips)
#define WIOS_24BIT_CLIP_COUNT(d,clips) WIOS_SIGNED_CLIP_COUNT(24,d,clips)
#define WIOS_32BIT_CLIP_COUNT(d,clips) WIOS_SIGNED_CLIP_COUNT(32,d,clips)

#define WIOS_CONVERT_SIGNED_TO_32BIT(bits,d) ((int32_t)(d)<<(32-bits))
#define WIOS_CONVERT_UNSIGNED_TO_32BIT(bits,d) (WIOS_CONVERT_SIGNED_TO_32BIT(bits,d)^WIOS_INT_MIN(32))

#define WIOS_CONVERT_U8_TO_32BIT(d) WIOS_CONVERT_UNSIGNED_TO_32BIT(8,d)
#define WIOS_CONVERT_16_TO_32BIT(d) WIOS_CONVERT_SIGNED_TO_32BIT(16,d)
#define WIOS_CONVERT_24_TO_32BIT(d) WIOS_CONVERT_SIGNED_TO_32BIT(24,d)

#define WIOS_RENEW_MIN_MAX_FRAME(d,mn,mx) do { (mn) = min((mn), d); (mx) = max((mx), d); } while (0)

#define CHANGE_USEC_TO_SEC(usec) ((usec)/(double)1e+6)
#define CHANGE_SEC_TO_USEC(sec) ((sec)*1e+6)

#define DEVICE_RUN_WARN(fmt,...) do { if (wios_warn_level > WIOS_WARN_LEVEL_MIN) {WIOS_WARN(fmt, __VA_ARGS__);} } while (0)

#define DA_CLIP_WARN(flag) do { if (flag && (wios_warn_level == WIOS_WARN_LEVEL_MAX)) {WIOS_WARN("Clipping occurs on output", NULL);} } while (0)
#define AD_CLIP_WARN(flag) do { if (flag && (wios_warn_level == WIOS_WARN_LEVEL_MAX)) {WIOS_WARN("Clipping occurs on input", NULL);} } while (0)
#endif /* WIOS_H_ */
